@@ -4,8 +4,8 @@ from __future__ import annotations
 
 from typing import TYPE_CHECKING
 
-from deebot_client.events import Position, PositionsEvent
 from deebot_client.message import HandlingResult
+from deebot_client.messages.xml import Pos
 from deebot_client.rs.map import PositionType
 
 from .common import XmlCommandWithMessageHandling
@@ -16,7 +16,7 @@ if TYPE_CHECKING:
     from deebot_client.event_bus import EventBus
 
 
-class GetPos(XmlCommandWithMessageHandling):
+class GetPos(XmlCommandWithMessageHandling, Pos):
     """GetPos command."""
 
     NAME = "GetPos"
@@ -30,13 +30,21 @@ class GetPos(XmlCommandWithMessageHandling):
         if xml.attrib.get("ret") != "ok" or xml.attrib.get("t") != "p":
             return HandlingResult.analyse()
 
-        if p := xml.attrib.get("p"):
-            p_x, p_y = p.split(",", 2)
-            p_a = xml.attrib.get("a", 0)
-            position = Position(
-                type=PositionType.DEEBOT, x=int(p_x), y=int(p_y), a=int(p_a)
-            )
-            event_bus.notify(PositionsEvent(positions=[position]))
-            return HandlingResult.success()
+        return cls._parse_xml(PositionType.DEEBOT, event_bus, xml)
 
-        return HandlingResult.analyse()
+
+class GetChargerPos(XmlCommandWithMessageHandling, Pos):
+    """GetChargerPos command."""
+
+    NAME = "GetChargerPos"
+
+    @classmethod
+    def _handle_xml(cls, event_bus: EventBus, xml: Element) -> HandlingResult:
+        """Handle xml message and notify the correct event subscribers.
+
+        :return: A message response
+        """
+        if xml.attrib.get("ret") != "ok":
+            return HandlingResult.analyse()
+
+        return cls._parse_xml(PositionType.CHARGER, event_bus, xml)
